package ssw.mj.test;

import java.io.StringReader;

import junit.framework.TestCase;

import ssw.mj.ErrorStrings;
import ssw.mj.Parser;
import ssw.mj.Scanner;
import ssw.mj.Token;
import ssw.mj.symtab.Obj;
import ssw.mj.symtab.Struct;
import ssw.mj.symtab.Tab;

/**
 * Test cases for the <code>Parser</code> class.
 * 
 * @author AW
 */
public class ParserTest extends TestCase {
	private static final String CR = "\r";
	private static final String LF = "\n";
	
	private static final int
		none       = Token.none,
		ident      = Token.ident,
		number     = Token.number,
		charConst  = Token.charConst,
		plus       = Token.plus,
		minus      = Token.minus,
		times      = Token.times,
		slash      = Token.slash,
		rem        = Token.rem,
		eql        = Token.eql,
		neq        = Token.neq,
		gtr        = Token.gtr,
		geq        = Token.geq,
		less       = Token.less,
		leq        = Token.leq,
		and        = Token.and,
		or         = Token.or,
		assign     = Token.assign,
		pplus      = Token.pplus,
		mminus     = Token.mminus,
		semicolon  = Token.semicolon,
		comma      = Token.comma,
		period     = Token.period,
		lpar       = Token.lpar,
		rpar       = Token.rpar,
		lbrack     = Token.lbrack,
		rbrack     = Token.rbrack,
		lbrace     = Token.lbrace,
		rbrace     = Token.rbrace,
		classKW    = Token.classKW,
		continueKW = Token.continueKW,
		doKW       = Token.doKW,
		elseKW     = Token.elseKW,
		finalKW    = Token.finalKW,
		ifKW       = Token.ifKW,
		newKW      = Token.newKW,
		printKW    = Token.printKW,
		readKW     = Token.readKW,
		returnKW   = Token.returnKW,
		voidKW     = Token.voidKW,
		whileKW    = Token.whileKW,
		eof        = Token.eof;

	private static final String
		//----- syntax errors
		CONST_DECL        = ErrorStrings.CONST_DECL,
		DESIGN_FOLLOW     = ErrorStrings.DESIGN_FOLLOW,
		INVALID_CHAR      = ErrorStrings.INVALID_CHAR,
		INVALID_FACT      = ErrorStrings.INVALID_FACT,
		INVALID_STAT      = ErrorStrings.INVALID_STAT,
		METH_DECL         = ErrorStrings.METH_DECL,
		NO_EOF            = ErrorStrings.NO_EOF,
		REL_OP            = ErrorStrings.REL_OP,
		//----- semantic errors
		CONST_TYPE    = ErrorStrings.CONST_TYPE,
		DECL_NAME     = ErrorStrings.DECL_NAME,
		INVALID_DECL  = ErrorStrings.INVALID_DECL,
		NO_ARRAY      = ErrorStrings.NO_ARRAY,
		NO_CLASS      = ErrorStrings.NO_CLASS,
		NO_CLASS_TYPE = ErrorStrings.NO_CLASS_TYPE,
		NO_FIELD      = ErrorStrings.NO_FIELD,
		NO_MAIN       = ErrorStrings.NO_MAIN,
		NO_PARAMS     = ErrorStrings.NO_PARAMS,
		NO_RETURN     = ErrorStrings.NO_RETURN,
		NO_TYPE       = ErrorStrings.NO_TYPE,
		NOT_FOUND     = ErrorStrings.NOT_FOUND,
		RETURN_VOID   = ErrorStrings.RETURN_VOID,
		VOID_MAIN     = ErrorStrings.VOID_MAIN;

	private static TestPrintWriter output;

	// ---------- helpers ----------
	
	private void initParser (String program) {
		output = new TestPrintWriter();
		Scanner.init(new StringReader(program), null);
	}

	// ----------------------------------------------------
	// ---------- test complete correct programs ----------

	/** Test the shortest syntactically valid program: */
	public void testShortestProgram () {
		initParser("class Shortest { }");

		Parser.parse(output);
		output.addExpectedLine("-- line 1, col 18 (SEM): " + NOT_FOUND + ": 'main'");
		output.verify();

		////////////////////////		
		// check symbol table //
		////////////////////////
		Obj actual, expected;

		// just to make sure that the Universe is also there		
		actual = Tab.find("eol");
		expected = new Obj(Obj.Con, "eol", Tab.charType, 10, 0, null);
		assertEquals("Universe 'eol'", expected, actual);

		// only name in symbol table of this program
		actual = Tab.find("Shortest");
		expected = new Obj(Obj.Prog, "Shortest", Tab.noType, 0, 0, null);
		assertEquals("Program 'Shortest'", expected, actual);

		// make sure that this is the end
		assertNull(actual.next);
	}

	public void testDoNothingProgram () {
		initParser(	"class DoNothing" + LF +
		            "\tint dummy;" + LF +
		            "{" + LF +
		            "\tint doNothing() { dummy=0; return dummy; }" + LF +
		            "" + LF +
		            "\tvoid main() { " + LF +
		            "\t\tdummy= doNothing();" + LF +
		            "\t}" + LF +
		            "}");

		Parser.parse(output);
		output.verify();
		
		////////////////////////		
		// check symbol table //
		////////////////////////
		Obj actual, expected;

		actual = Tab.find("DoNothing");
		assertNotNull(actual);
		expected = new Obj(Obj.Prog, "DoNothing", Tab.noType, 0, 0, null);
		// we cheat here, because we will compare all the rest later
		expected.locals = actual.locals;
		assertEquals("Program 'DoNothing'", expected, actual);
		
		// Knowing the implementation of the symbol table
		// we can traverse it and check all the names separately
		// NOTE: thus we get better error messages
		
		actual = actual.locals;
		expected = new Obj(Obj.Var, "dummy", Tab.intType, 0, 0, null);
		assertEquals("Global 'dummy'", expected, actual);
		
		actual = actual.next;
		expected = new Obj(Obj.Meth, "doNothing", Tab.intType, 0, 0, null);
		assertEquals("Method 'doNothing'", expected, actual);
		
		actual = actual.next;
		expected = new Obj(Obj.Meth, "main", Tab.noType, 0, 0, null);
		assertEquals("Method 'main'", expected, actual);
		
		// make sure that this is the end
		assertNull(actual.next);
	}

	public void testScriptExample () {
		initParser( "class P" + LF +
		            "" + LF +
		            "  final int size = 10;" + LF +
		            "" + LF +
		/*line  5*/ "  class Table {" + LF +
		            "    int[] pos;" + LF +
		            "    int[] neg;" + LF +
		            "  }" + LF +
		            "" + LF +
		/*line 10*/ "  Table val;" + LF +
		            "{" + LF +
		            "  void main()" + LF +
		            "    int x, i;" + LF +
		            "  { /*---------- Initialize val */" + LF +
		/*line 15*/ "    val = new Table;" + LF +
		            "    val.pos = new int[size]; val.neg = new int[size];" + LF +
		            "    i = 0;" + LF +
		            "    while (i < size) {" + LF +
		            "      val.pos[i] = 0; val.neg[i] = 0;" + LF +
		/*line 20*/ "      i++;" + LF +
		            "    }" + LF +
		            "    /*---------- Read values */" + LF +
		            "    read(x);" + LF +
		            "    while (x > 0) {" + LF +
		/*line 25*/ "      if (0 <= x && x < size) {" + LF +
		            "        val.pos[x]++;" + LF +
		            "      } else if (-size < x && x < 0) {" + LF +
		            "        val.neg[-x]++;" + LF +
		            "      }" + LF +
		/*line 30*/ "      read(x);" + LF +
		            "    }" + LF +
		            "  }" + LF +
		            "}");
		
		Parser.parse(output);
		output.verify();
		
		////////////////////////
		// check symbol table //
		////////////////////////
		Obj actual, expected;

		actual = Tab.find("P");
		assertNotNull(actual);
		expected = new Obj(Obj.Prog, "P", Tab.noType, 0, 0, null);
		// We cheat here, because we will compare all the rest later.
		expected.locals = actual.locals;
		assertEquals("Program 'P'", expected, actual);
		
		// Knowing the implementation of the symbol table
		// we can traverse it and check all the names separately.
		// NOTE: This gives us more precise error messages
		//       than just re-creating the whole symbol table and
		//       have it check by one call to equals(Prog).
		
		actual = actual.locals;
		expected = new Obj(Obj.Con, "size", Tab.intType, 10, 0, null);
		assertEquals("Constant 'size'", expected, actual);
		
		actual = actual.next;
		Struct type = new Struct(Struct.Class); 
		type.n = 2; 
		// No cheating here, field list is small enough.
		// So we explicitly re-create it and check it at once (with equals).
		type.fields = new Obj(Obj.Fld, "pos", 
		                      new Struct(Struct.Arr, Tab.intType), 0, 1, null);
		assertEquals("Field 'pos'", type.fields, actual.type.fields);
		type.fields.next = new Obj(Obj.Fld, "neg", 
		                           new Struct(Struct.Arr, Tab.intType), 1, 1, null);
		assertEquals("Field 'neg'", type.fields.next, actual.type.fields.next);
		expected = new Obj(Obj.Type, "Table", type, 0, 0, null);
		assertEquals("Type 'Table'", expected, actual);
		
		actual = actual.next;
		expected = new Obj(Obj.Var, "val", type, 0, 0, null);
		assertEquals("Global 'val'", expected, actual);
		
		actual = actual.next;
		Obj local = new Obj(Obj.Var, "x", Tab.intType, 0, 1, null);
		assertEquals("Local 'x'", local, actual.locals);
		local.next = new Obj(Obj.Var, "i", Tab.intType, 1, 1, null);
		assertEquals("Local 'i'", local.next, actual.locals.next);
		expected = new Obj(Obj.Meth, "main", Tab.noType, 0, 0, local);
		assertEquals("Method 'main'", expected, actual);
		
		// make sure that this is the end
		assertNull(actual.next);
	}
	
	public void testAllProds () {
		initParser(	"class AllProds" + LF +
		            "  final int nPoints = 2;" + LF +
		            "  final char exclaim = '!';" + LF +
		            "" + LF +
		/*line  5*/	"  class Point {" + LF +
		            "    int x, y;" + LF +
		            "  }" + LF +
		            "" + LF +
		            "  char[] msg;" + LF +
		/*line 10*/	"  Point[] pArr;" + LF +
		            "" + LF +
		            "{" + LF +
		            "  void printError ()" + LF +
		            "    int i;" + LF +
		/*line 15*/	"  {" + LF +
		            "    /* print error message */" + LF +
		            "    i = 0;" + LF +
		            "    while(i<len(msg)){print(msg[i]);i++;}" + LF +
		            "    print(exclaim);" + LF +
		/*line 20*/	"    do {" + LF +
		            "      i--;" + LF +
		            "      if (i >= 5) continue;" + LF +
		            "      print(msg[i], 2);" + LF +
		            "    } while(i > 0);" + LF +
		/*line 25*/	"  }" + LF +
		            "" + LF +
		            "  int getX (Point p) { return p.x; }" + LF +
		            "" + LF +
		            "  void setX (int x, Point p) {" + LF +
		/*line 30*/	"    if (x > -10 && x <= 9 ||" + LF +
		            "        x == 4*3/4%4+12-11*(2147483647/(-2147483647-1)%2) )" + LF +
		            "      p.x = x;" + LF +
		            "    else printError();" + LF +
		            "  }" + LF +
		/*line 35*/	"" + LF +
		            "" + LF +
		            "  void setMsg () {" + LF +
		            "    msg = new char[5];" + LF +
		            "    msg[0] = 'E';" + LF +
		/*line 40*/	"    msg[1] = 'r';" + LF +
		            "    msg[2] = 'r';" + LF +
		            "    msg[3] = 'o';" + LF +
		            "    msg[4] = 'r';" + LF +
		            "  }" + LF +
		/*line 45*/	"" + LF +
		            "  void setPoints ()" + LF +
		            "    int i;" + LF +
		            "  {" + LF +
		            "    pArr = new Point[nPoints];" + LF +
		/*line 50*/	"    i = 0;" + LF +
		            "    while (i < len(pArr)) {" + LF +
		            "      pArr[i] = new Point;" + LF +
		            "      pArr[i].x = i;" + LF +
		            "      pArr[i].y = i;" + LF +
		/*line 55*/	"      i++;" + LF +
		            "    }" + LF +
		            "  }" + LF +
		            "" + LF +
		            "  void main ()" + LF +
		/*line 60*/	"    int i;" + LF +
		            "  {" + LF +
		            "    setMsg();" + LF +
		            "    setPoints();" + LF +
		            "    read(i);" + LF +
		/*line 65*/	"    if (i < nPoints) {" + LF +
		            "      print(getX(pArr[i]));" + LF +
		            "      print(pArr[i].y);" + LF +
		            "    }" + LF +
		            "    else" + LF +
		/*line 70*/	"      printError();" + LF +
		            "    print(eol);" + LF +
		            "  }" + LF +
		            "}");

		Parser.parse(output);
		output.verify();

		////////////////////////		
		// check symbol table //
		////////////////////////

		Obj actual, expected;

		actual = Tab.find("AllProds");
		assertNotNull(actual);
		expected = new Obj(Obj.Prog, "AllProds", Tab.noType, 0, 0, null);
		// We cheat here, because we will compare all the rest later.
		expected.locals = actual.locals;
		assertEquals("Program 'AllProds'", expected, actual);
		
		actual = actual.locals;
		expected = new Obj(Obj.Con, "nPoints", Tab.intType, 2, 0, null);
		assertEquals("Constant 'nPoints'", expected, actual);
		
		actual = actual.next;
		expected = new Obj(Obj.Con, "exclaim", Tab.charType, 33, 0, null);
		assertEquals("Constant 'exclaim'", expected, actual);
		
		actual = actual.next;

		Struct pointT = new Struct(Struct.Class); 
		pointT.n = 2; 
		pointT.fields = new Obj(Obj.Fld, "x", Tab.intType, 0, 1, null);
		assertEquals("Field 'x'", pointT.fields, actual.type.fields);
		pointT.fields.next = new Obj(Obj.Fld, "y", Tab.intType, 1, 1, null);
		assertEquals("Field 'y'", pointT.fields.next, actual.type.fields.next);

		expected = new Obj(Obj.Type, "Point", pointT, 0, 0, null);
		assertEquals("Class 'Point'", expected, actual);
		
		actual = actual.next;
		expected = new Obj(Obj.Var, "msg", 
		                   new Struct(Struct.Arr, Tab.charType), 0, 0, null);
		assertEquals("Global 'msg'", expected, actual);

		actual = actual.next;
		expected = new Obj(Obj.Var, "pArr", 
							new Struct(Struct.Arr, pointT), 1, 0, null);
		assertEquals("Global 'pArr'", expected, actual);

		actual = actual.next;	// printError
		Obj local = new Obj(Obj.Var, "i", Tab.intType, 0, 1, null);
		assertEquals("Local 'i'", local, actual.locals);
		expected = new Obj(Obj.Meth, "printError", Tab.noType, 0, 0, local);
		assertEquals("Method 'printError'", expected, actual);
		
		actual = actual.next;	// getX
		local = new Obj(Obj.Var, "p", pointT, 0, 1, null);
		assertEquals("Local 'p'", local, actual.locals);
		expected = new Obj(Obj.Meth, "getX", Tab.intType, 0, 1, local);
		assertEquals("Method 'getX'", expected, actual);
		
		actual = actual.next; 	// setX
		local = new Obj(Obj.Var, "x", Tab.intType, 0, 1, null);
		assertEquals("Local 'x'", local, actual.locals);
		local.next = new Obj(Obj.Var, "p", pointT, 1, 1, null);
		assertEquals("Local 'p'", local.next, actual.locals.next);
		expected = new Obj(Obj.Meth, "setX", Tab.noType, 0, 2, local);
		assertEquals("Method 'setX'", expected, actual);
		
		actual = actual.next;	// setMsg
		expected = new Obj(Obj.Meth, "setMsg", Tab.noType, 0, 0, null);
		assertEquals("Method 'setMsg'", expected, actual);
		
		actual = actual.next;	// setPoints
		local = new Obj(Obj.Var, "i", Tab.intType, 0, 1, null);
		assertEquals("Local 'i'", local, actual.locals);
		expected = new Obj(Obj.Meth, "setPoints", Tab.noType, 0, 0, local);
		assertEquals("Method 'setPoints'", expected, actual);
		
		actual = actual.next;	// main
		// local is same as above
		assertEquals("Local 'i'", local, actual.locals);
		expected = new Obj(Obj.Meth, "main", Tab.noType, 0, 0, local);
		assertEquals("Method 'main'", expected, actual);
		
		// make sure that this is the end
		assertNull(actual.next);
	}
	
	// -----------------------------------------------
	// ---------- test programs with errors ----------

	public void testScanErrIn1stSymbol () {
		initParser("");	
		
		Parser.parse(output);
		output.addExpectedLine("-- line 1, col 1 (SYN): " + INVALID_CHAR + " ''");
		output.verify();
	}
	
	public void testWrongStart () {
		initParser("noclass SomeClass { }");
		
		Parser.parse(output);
		output.addExpectedLine("-- line 1, col 1 (SYN): " + Token.names[classKW] + 
		                       " expected");	
		output.verify();
	}
	
	public void testNoProgName () {
		initParser(	"class { }");
		
		Parser.parse(output);
		output.addExpectedLine("-- line 1, col 7 (SYN): " + Token.names[ident] + " expected");
		output.verify();	
	}

	public void testUndefName () {
		initParser("class SSW" + LF +
		           "  integer x;" + LF +
		           "{ void main () { } }");
		
		Parser.parse(output);
		output.addExpectedLine("-- line 2, col 11 (SEM): " + NOT_FOUND + ": 'integer'");
		output.verify();	
	}
	
	public void testWrongConstType () {
		initParser("class SSW" + LF +
		           "  final char ch = 32;" + LF +
		           "{ void main () { } }");

		Parser.parse(output);
		output.addExpectedLine("-- line 2, col 19 (SEM): " + CONST_TYPE);
		output.verify();			
	}

	public void testWrongGlobalDecl () {
		initParser("class SSW" + LF +
		           "  final int MAX = 10;" + LF +
		           "  123, i, x;" + LF +
		           "  char c1, c2, c3;" + LF +
		           "{ void main () { } }");

		Parser.parse(output);
		output.addExpectedLine("-- line 3, col 3 (SYN): " + INVALID_DECL);
		output.verify();	

		////////////////////////		
		// check symbol table //
		////////////////////////

		Obj actual, expected;

		actual = Tab.find("SSW");
		expected = new Obj(Obj.Prog, "SSW", Tab.noType, 0, 0, null);
		// We cheat here, because we will compare all the rest later.
		expected.locals = actual.locals;
		assertEquals("Program 'SSW'", expected, actual);
		
		actual = actual.locals;  // MAX
		expected = new Obj(Obj.Con, "MAX", Tab.intType, 10, 0, null);
		assertEquals("Constant 'MAX'", expected, actual);
		
		actual = actual.next;  // c1
		expected = new Obj(Obj.Var, "c1", Tab.charType, 0, 0, null);
		assertEquals("Global 'c1'", expected, actual);
		
		actual = actual.next;  // c2
		expected = new Obj(Obj.Var, "c2", Tab.charType, 1, 0, null);
		assertEquals("Global 'c2'", expected, actual);

		actual = actual.next;  // c3
		expected = new Obj(Obj.Var, "c3", Tab.charType, 2, 0, null);
		assertEquals("Global 'c3'", expected, actual);

		actual = actual.next;  // main
		expected = new Obj(Obj.Meth, "main", Tab.noType, 0, 0, null);
		assertEquals("Method 'main'", expected, actual);

		// make sure that this is the end
		assertNull(actual.next);
	}

	public void testWrongStat () {
		initParser("class SSW {" + LF +
		           "  void main () {  " + LF +
		           "    if (0 > 2) ;" + LF +
		           "    123;" + LF +
		           "    while (5 < 6) continue;" + LF +
		           "    567 no start on ident;" + LF +
		           "  }" + LF +
		           "}");
		
		Parser.parse(output);
		output.addExpectedLine("-- line 4, col 5 (SYN): " + INVALID_STAT);
		output.addExpectedLine("-- line 6, col 5 (SYN): " + INVALID_STAT);
		output.verify();
	}

	public void testNoClassType () {
		initParser("class SSW" + LF +
		           "  class Point { int x, y; }" + LF +
		           "{ void main () Point p; {" + LF +
		           "    p = new Point;" + LF +
		           "    p = new int;" + LF +
		           "  }" + LF +
		           "}");

		Parser.parse(output);
		output.addExpectedLine("-- line 5, col 16 (SEM): " + NO_CLASS_TYPE);
		output.verify();
	}

	public void testMainVar () {
		initParser("class SSW" + LF +
		           "  int main;" + LF +
		           "{ }");

		Parser.parse(output);
		output.addExpectedLine("-- line 3, col 3 (SEM): " + NO_MAIN);
		output.verify();
	}

	public void testMainWithParams () {
		initParser("class SSW {" + LF +
		           "  void main (int i) { }" + LF +
		           "}");

		Parser.parse(output);
		output.addExpectedLine("-- line 2, col 21 (SEM): " + NO_PARAMS);
		output.verify();
	}

	public void testNoReturnVal () {
		initParser("class SSW {" + LF +
		           "  void main () {}" + LF +
		           "  int getX () { return; }" + LF +
		           "}");
		Parser.parse(output);
		output.addExpectedLine("-- line 3, col 23 (SEM): " + NO_RETURN);
		output.verify();
	}

	public void testNoType () {
		initParser("class SSW" + LF +
		           "  final int S = 10;" + LF +
		           "  S s;" + LF +
		           "  class Point { int x, y; }" + LF +
		           "{" + LF +
		           "  void main () Point p; {" + LF +
		           "    p = new S;" + LF +
		           "  }" + LF +
		           "}");

		Parser.parse(output);
		output.addExpectedLine("-- line 3, col 5 (SEM): " + NO_TYPE);
		output.addExpectedLine("-- line 7, col 14 (SEM): " + NO_TYPE);
		output.verify();
	}

	public void testWrongReturnVal () {
		initParser("class SSW {" + LF +
		           "  void main () {}" + LF +
		           "  void doIt () { return 5; }" + LF +
		           "}");
		Parser.parse(output);
		output.addExpectedLine("-- line 3, col 25 (SEM): " + RETURN_VOID);
		output.verify();
	}

	public void testMainNotVoid () {
		initParser("class SSW {" + LF +
		           "  char main () { }" + LF +
		           "}");

		Parser.parse(output);
		output.addExpectedLine("-- line 2, col 16 (SEM): " + VOID_MAIN);
		output.verify();		
	}

	public void testDoubleDecl () {
		initParser("class SSW " + LF +
		           "  int x, x;" + LF +
		           "{" + LF +
		           "  void main () int o; char c, o; { }" + LF +
		           "}");
		
		Parser.parse(output);
		output.addExpectedLine("-- line 2, col 11 (SEM): " + DECL_NAME + ": 'x'");
		output.addExpectedLine("-- line 4, col 32 (SEM): " + DECL_NAME + ": 'o'");
		output.verify();
	}

	public void testNoField () {
		initParser("class SSW" + LF +
		           "  class Point { int x, y; }" + LF +
		           "{" + LF +
		           "  void main () Point p; {" + LF +
		           "    print(p.x);" + LF +
		           "    print(p.z);" + LF +
		           "  }" + LF +
		           "}");
		
		Parser.parse(output);
		output.addExpectedLine("-- line 6, col 14 (SEM): " + NO_FIELD + ": 'z'");
		output.verify();
	}

	public void testNoArray () {
		initParser("class SSW " + LF +
		           "  int x;" + LF +
		           "{" + LF +
		           "  void main () {" + LF +
		           "    x[1]++;" + LF +
		           "  }" + LF +
		           "}");
		
		Parser.parse(output);
		output.addExpectedLine("-- line 5, col 9 (SEM): " + NO_ARRAY);
		output.verify();
	}
	
	public void testAllProdsWithErrors () {
		initParser(	"class AllProdsWithErros" + LF +
		            "  final int nPoints = 2;" + LF +
		            "  final exclaim = '!';" + LF +
		            "" + LF +
		/*line  5*/ "  class {" + LF +
		            "    int x, y;" + LF +
		            "  }" + LF +
		            "" + LF +
		            "  char[ msg;" + LF +
		/*line 10*/ "  Point[] pArr;" + LF +
		            "" + LF +
		            "{" + LF +
		            "  void printError ()" + LF +
		            "    int i;" + LF +
		/*line 15*/ "  {" + LF +
		            "    /* print error message */" + LF +
		            "    i = 0;" + LF +
		            "    while(i<len(msg)){print(msg[i]);i++;}" + LF +
		            "    print(exclaim);" + LF +
		/*line 20*/ "    do {" + LF +
		            "      i--;" + LF +
		            "      if (i >= 5) continue;" + LF +
		            "      print(msg[i]);" + LF +
		            "    } while(i > 0);" + LF +
		/*line 25*/ "  }" + LF +
		            "" + LF +
		            "  int getX (Point p) { return p.x; }" + LF +
		            "" + LF +
		            "  void setX (int x, Point p) {" + LF +
		/*line 30*/ "    if (x > -10 && x <= 9 ||" + LF +
		            "        x == 4*3/4%4+12-11*(2147483647/(-2147483647-1)%2) )" + LF +
		            "      p.x = x;" + LF +
		            "    else printError();" + LF +
		            "  }" + LF +
		/*line 35*/ "" + LF +
		            "" + LF +
		            "  void setMsg () {" + LF +
		            "    msg = new char[6];" + LF +
		            "    msg[0] 'E';" + LF +
		/*line 40*/ "    msg[1] = 'r';" + LF +
		            "    msg[2] = 'r';" + LF +
		            "    msg[3] = 'o';" + LF +
		            "    msg[4] = 'r';" + LF +
		            "    msg[5] = '\\n';" + LF +
		/*line 45*/ "  }" + LF +
		            "" + LF +
		            "  void setPoints ()" + LF +
		            "    int i;" + LF +
		            "  {" + LF +
		/*line 50*/ "    pArr = new Point[nPoints];" + LF +
		            "    i = 0;" + LF +
		            "    while (i < len(pArr)) {" + LF +
		            "      pArr[i] = new Point;" + LF +
		            "      pArr[i].x = i;" + LF +
		/*line 55*/ "      pArr[i].y = i" + LF +
		            "      i++;" + LF +
		            "    }" + LF +
		            "  }" + LF +
		            "" + LF +
		/*line 60*/ "  int main ()" + LF +
		            "    int i;" + LF +
		            "  {" + LF +
		            "    setMsg();" + LF +
		            "    setPoints();" + LF +
		/*line 65*/ "    read(i);" + LF +
		            "    if (i < nPoints) {" + LF +
		            "      print(getX(pArr[i]));" + LF +
		            "      print(pArr[i].y);" + LF +
		            "    }" + LF +
		/*line 70*/ "    else" + LF +
		            "      printError();" + LF +
		            "    print(eol);" + LF +
		            "  }" + LF +
		            "}");

		Parser.parse(output);
		output.addExpectedLine("-- line 3, col 17 (SEM): " + NOT_FOUND + ": 'exclaim'");
		output.addExpectedLine("-- line 5, col 9 (SYN): " + Token.names[ident] + " expected");
		output.addExpectedLine("-- line 9, col 9 (SYN): " + Token.names[rbrack] + 
		                       " expected");
		output.addExpectedLine("-- line 10, col 8 (SEM): " + NOT_FOUND + ": 'Point'");
		output.addExpectedLine("-- line 27, col 19 (SEM): " + NOT_FOUND + ": 'Point'");
		output.addExpectedLine("-- line 27, col 34 (SEM): " + NO_CLASS);
		output.addExpectedLine("-- line 29, col 27 (SEM): " + NOT_FOUND + ": 'Point'");
		output.addExpectedLine("-- line 32, col 11 (SEM): " + NO_CLASS);
		output.addExpectedLine("-- line 39, col 12 (SYN): " + DESIGN_FOLLOW);
		output.addExpectedLine("-- line 50, col 21 (SEM): " + NOT_FOUND + ": 'Point'");
		output.addExpectedLine("-- line 53, col 26 (SEM): " + NOT_FOUND + ": 'Point'");
		output.addExpectedLine("-- line 54, col 17 (SEM): " + NO_CLASS);
		output.addExpectedLine("-- line 55, col 17 (SEM): " + NO_CLASS);
		output.addExpectedLine("-- line 61, col 5 (SEM): " + VOID_MAIN);
		output.addExpectedLine("-- line 68, col 22 (SEM): " + NO_CLASS);
		output.verify();
	}

	public void testNoEof () {
		initParser( "class X { void main () {} } more text");
		
		Parser.parse(output); 
		output.addExpectedLine("-- line 1, col 29 (SYN): " + NO_EOF);
		output.verify();				
	}

	// -----------------------------------------------
				
	public static void main (String[] args) {
		junit.textui.TestRunner.run(ParserTest.class);
	}
}
